include	'../include/padAMD64.inc'

align 16
bp_transform_address2index:
; input		RDX	breakpoint address
; output	R12	breakpoint index, which means:
;			N breakpoint address is at qword [BreakpointAddress+8*N]
;		set or clear carry flag depending in success or failure
;		write message in case of failure

; find if there is a breakpoint at the address:
	lea	rax,[rdx]
	mov	ecx,MaxNumberOfBreakpoints
	lea	r12,[rcx-1]
	lea	rdi,[BreakpointsAddresses]
	repnz scasq
	lea	rax,[msg_bp_not_exists]
	jnz	manipulate_bp_failure
	sub	r12,rcx			; resulting breakpoint index
;	clc
; clc isn't necessary because above sub clears carry flag everytime
	ret

align 16
bp_transform_address2index_put_bp_into_slot:
; input		RDX	breakpoint address
; output	R12	breakpoint index, which means:
;			N breakpoint address is at qword [BreakpointAddress+8*N]
;		put RDX into [BreakpointsAddresses + r12*8] if success
;		set or clear carry flag depending in success or failure
;		write message in case of failure

; find if there isn't any breakpoint at the same address:
	lea	rax,[rdx]
	mov	ecx,MaxNumberOfBreakpoints
	lea	rdi,[BreakpointsAddresses]
	repnz scasq
	lea	rax,[msg_bp_exists]
	jz	manipulate_bp_failure

; find first empty slot:
	xor	eax,eax			; empty slot holds qword 0
	mov	ecx,MaxNumberOfBreakpoints
	lea	r12,[rcx-1]
	lea	rdi,[BreakpointsAddresses]
	repnz scasq
	lea	rax,[msg_max_breakpoints_set]
	jnz	manipulate_bp_failure
	sub	r12,rcx			; resulting breakpoint index
	lea	rdi,[BreakpointsAddresses]
	mov	qword [rdi+r12*8],rdx
;	clc
; clc isn't necessary because above sub clears carry flag everytime
	ret

align 16
put_bp:
; input		R12	breakpoint index
; output	set or clear carry flag depending success or failure
	lea	rax,[BreakpointsAddresses]
	lea	r10,[code_buf]		; destination data buffer in fdbg
	mov	rdx,qword [rax + r12*8]	; source address in program
	lea	rdi,[msg_error_peektext_address]
	lea	rax,[rdx]
	call	hexa_64_rax
	lea	rax,[msg_abnormal_address]
	or	rdx,rdx
	jz	manipulate_bp_failure
	mov	rsi,qword [child_PID]
	mov	edi,PTRACE_PEEKTEXT
	mov	eax,sys_ptrace
	syscall
	or	rax,rax
	lea	rax,[msg_error_peektext]
	js	manipulate_bp_failure

	mov	r10,qword [code_buf]
	lea	rdx,[BreakpointOldByte]
	lea	rax,[msg_bp_exists_in_code]
	cmp	r10b,0CCh			; instruction of int 03
						; the second way is word CD 03
	jz	manipulate_bp_failure
	cmp	r10w,03CDh			; warning, dword operation
						; like cmp r10d,... clears upper
						; dword of r10 !!!!!
						; word operation doesn't zeroing
	jz	manipulate_bp_failure
	mov	byte [rdx + r12*1],r10b		; save old byte at bp position
	mov	r10b,0CCh			; prepare int3 to be written

	lea	rax,[BreakpointsAddresses]
; r10 holds now new value to be written - 4th param for syscall
	mov	rdx,qword [rax + r12*8]	; destination address in program
	lea	rdi,[msg_error_poketext_address]
	lea	rax,[rdx]
	call	hexa_64_rax
	mov	rsi,qword [child_PID]
	mov	edi,PTRACE_POKETEXT
	mov	eax,sys_ptrace
	syscall
	or	rax,rax
	lea	rax,[msg_error_poketext]
	js	manipulate_bp_failure
	clc
	ret

manipulate_bp_failure:
	call	wzts_std
	stc
	ret

align 16
remove_bp:
; input		R12	breakpoint index
; output	set or clear carry flag depending success or failure
;		call purge_bp_slot when success
	lea	rax,[BreakpointsAddresses]
	lea	r10,[code_buf]		; destination data buffer in fdbg
	mov	rdx,qword [rax + r12*8]	; source address in program
	lea	rdi,[msg_error_peektext_address]
	lea	rax,[rdx]
	call	hexa_64_rax
	lea	rax,[msg_abnormal_address]
	or	rdx,rdx
	jz	manipulate_bp_failure
	mov	rsi,qword [child_PID]
	mov	edi,PTRACE_PEEKTEXT
	mov	eax,sys_ptrace
	syscall
	or	rax,rax
	lea	rax,[msg_error_peektext]
	js	manipulate_bp_failure

	mov	r10,qword [code_buf]		; source from data buf in fdbg
	lea	rdx,[BreakpointOldByte]
	lea	rax,[msg_bp_not_exists]
	cmp	r10b,0CCh			; instruction of int 03
	jnz	manipulate_bp_failure
	mov	r10b,byte [rdx + r12*1]		; get saved old byte at bp

	lea	rax,[BreakpointsAddresses]
; r10 holds now new value to be written - 4th param for syscall
	mov	rdx,qword [rax + r12*8]	; destination address in program
	lea	rdi,[msg_error_poketext_address]
	lea	rax,[rdx]
	call	hexa_64_rax
	mov	rsi,qword [child_PID]
	mov	edi,PTRACE_POKETEXT
	mov	eax,sys_ptrace
	syscall
	or	rax,rax
	lea	rax,[msg_error_poketext]
	js	manipulate_bp_failure

	call	purge_bp_slot
	clc
; clc isn't necessary because above subtraction in call purge_bp_slot clears
; carry flag everytime
	ret

align 16
purge_bp_slot:
; input	R12 slot index
	sub	ecx,ecx
	lea	rax,[BreakpointsAddresses]
	lea	rdx,[BreakpointOldByte]
	mov	qword [rax + r12*8],rcx	; clear breakpoint address
	mov	byte [rdx + r12*1],cl	; clear breakpoint saved old byte
	ret

align 16
patch_breakpoints_in_displayed_code_mode0:
; input	RBP address of code in program
;	code_buf holds code read from program
;	BreakpointsAddresses
;	BreakpointOldByte
; fast for big code_buf, slow for big MaxNumberOfBreakpoints
	lea	r10,[code_buf]
	lea	r8,[rbp + code_buf_size]	; end of code_buf
	lea	rsi,[BreakpointsAddresses]
	mov	ecx,MaxNumberOfBreakpoints
AMDPad16
patch_breakpoints_in_displayed_code_mode0_L0:
	lodsq
; this may make routine faster if MaxNumberOfBreakpoints is big:
; but make it slow if much of breakpoints and if breakpoints looks like:
; breakpoint0, dq 0, breakpoint1, dq 0, dq 0, breakpoint2, dq 0, breakpoint3,...
	or	rax,rax
	jnz	patch_breakpoints_in_displayed_code_mode0_L1
	lea	rdi,[rsi]
	dec	ecx
	repz scasq
	lea	ecx,[ecx+1]			; this don't change zero flag
						; instead of inc ecx
	jz	patch_breakpoints_in_displayed_code_mode0_L3
	mov	rax,qword [rdi-8]
	lea	rsi,[rdi]

patch_breakpoints_in_displayed_code_mode0_L1:
; optimalization for big MaxNumberOfBreakpoints done.
;

	cmp	rbp,rax
	jnbe	patch_breakpoints_in_displayed_code_mode0_L2
	cmp	r8,rax
	jbe	patch_breakpoints_in_displayed_code_mode0_L2
; patch it !
	mov	edx,MaxNumberOfBreakpoints
	sub	edx,ecx
	sub	rax,rbp
	lea	rdi,[BreakpointOldByte]
	mov	dl,byte [rdi + rdx*1]
	mov	byte [r10+rax],dl

patch_breakpoints_in_displayed_code_mode0_L2:
	loop	patch_breakpoints_in_displayed_code_mode0_L0
patch_breakpoints_in_displayed_code_mode0_L3:
	ret

align 16
patch_breakpoints_in_displayed_code_mode1:
; input	RBP address of code in program
;	code_buf holds code read from program
;	BreakpointsAddresses
;	BreakpointOldByte
	lea	r10,[code_buf]
	lea	rax,[rbp]
	mov	rbx,code_buf_size

AMDPad16
patch_breakpoints_in_displayed_code_mode1_L0:
	cmp	byte [r10],0CCh
	jnz	patch_breakpoints_in_displayed_code_mode1_L5
	mov	edx,MaxNumberOfBreakpoints - 1
	mov	ecx,MaxNumberOfBreakpoints
	lea	rdi,[BreakpointsAddresses]
	repnz scasq
	jnz	patch_breakpoints_in_displayed_code_mode1_L5
	sub	rdx,rcx
; patch it !
	lea	rdi,[BreakpointOldByte]
	mov	cl,byte [rdi + rdx*1]
	mov	byte [r10],cl
patch_breakpoints_in_displayed_code_mode1_L5:
	inc	r10
	inc	rax
	dec	rbx
	jnz	patch_breakpoints_in_displayed_code_mode1_L0
	ret

align 16
patch_breakpoints_before_writing_memory:
; in:	rax address
;	ecx size
;	rdi buffer
; destroys rdx
	push	rcx

patch_breakpoints_before_writing_memory_L0:
	push	rcx
	push	rdi
	lea	rdi,[BreakpointsAddresses]
	mov	ecx,MaxNumberOfBreakpoints
	lea	edx,[rcx-1]
	repne scasq
	pop	rdi
	jnz	patch_breakpoints_before_writing_memory_L5
	sub	edx,ecx
	push	rdi
	mov	cl,0CCh
	xchg	cl,[rdi]
	lea	rdi,[BreakpointOldByte]
	mov	[rdi+rdx],cl
	pop	rdi

patch_breakpoints_before_writing_memory_L5:
	pop	rcx
	scasb
	inc	rax
	loop	patch_breakpoints_before_writing_memory_L0
	pop	rcx
	ret

align 16
patch_breakpoints_before_write_code:
; in:	rax address
;	ebx size
;	rdi buffer
; destroys nonvolatile registers
	mov	r11d,MaxNumberOfBreakpoints-1
	lea	r10,[BreakpointsAddresses]
	lea	r9,[rax+rbx*1]
AMDPad16
patch_breakpoints_before_write_code_L0:
	mov	rcx,[r10+r11*8]
	cmp	rax,rcx
	jnbe	patch_breakpoints_before_write_code_L1
	cmp	r9,rcx
	jbe	patch_breakpoints_before_write_code_L1
	sub	rcx,rax
	mov	dl,[rdi+rcx*1]
	mov	[r10+8*MaxNumberOfBreakpoints+r11*1],dl	; set BreakpointOldByte
	mov	byte [rdi+rcx*1],0CCh

patch_breakpoints_before_write_code_L1:
	dec	r11d
	jns	patch_breakpoints_before_write_code_L0
	ret

align 16
patch_breakpoints_after_read_code:
; in:	rax address
;	ebx size
;	rdi buffer
; destroys nonvolatile registers
	mov	r11d,MaxNumberOfBreakpoints-1
	lea	r10,[BreakpointsAddresses]
	lea	r9,[rax+rbx*1]
AMDPad16
patch_breakpoints_after_read_code_L0:
	mov	rcx,[r10+r11*8]
	cmp	rax,rcx
	jnbe	patch_breakpoints_after_read_code_L1
	cmp	r9,rcx
	jbe	patch_breakpoints_after_read_code_L1
	sub	rcx,rax
	mov	dl,[r10+8*MaxNumberOfBreakpoints+r11*1]	; get BreakpointOldByte
	mov	[rdi+rcx*1],dl

patch_breakpoints_after_read_code_L1:
	dec	r11d
	jns	patch_breakpoints_after_read_code_L0
	ret
