align 16
convert_st0_32:					; 32 bits
convert_st0_sp:					; = single precision
	mov	dx,2*100h+8			; display 8 mantissa digits and 2 exponent digits
	jmp	convert_st0

align 16
convert_st0_64:					; 64 bits
convert_st0_dp:					; = double precision
	mov	dx,3*100h+16			; display 16 mantissa digits and 3 exponent digits
	jmp	convert_st0

align 16
convert_st0_80:
	mov	dx,4*100h+18			; display 18 mantissa digits and 4 exponent digits
;	jmp	convert_st0

align 16
convert_st0:
; input:	st0 = number
;		dl number of mantissa digits to display, must be parity number
;		dh number of exponent digits to display, must be 2,3,4
;		rdi point to output string buffer, must have 4 qwords size (18+4+4+3+1=30 bytes)
;		18 mantissa, 4 exponent, 4 '*10^', 3 mantissa sign + decimal point + exponent sign, 1 zero terminator
	push	rax rcx rdx r10 r11 r12 r13 r14 r15
	movzx	eax,dh				; exponent size
	movzx	ecx,dl				; mantissa size
	lea	eax,[eax+ecx]			; exponent size + mantissa size
	lea	rcx,[rdi+rax+4+3]		; rdi + exponent size + mantissa size + 4 ; 4 for *10^ ; 3 for 2 signs + 1 decimal point
	push	rcx				; save end rdi position on return
	mov	rax,'        '
	mov	qword [rdi+8*0],rax
	mov	qword [rdi+8*1],rax
	mov	qword [rdi+8*2],rax
	mov	qword [rdi+8*3],rax
	fxam					; check number type
	fstsw	[rdi]				; store status
	mov	eax,dword [rdi]
	bt	eax,8
	jc	convert_st0_specials
	call	convert_float_st0		; display decimal number
convert_st0_L9:
	pop	rdi
	mov	byte [rdi],0			; zero-terminated string
	pop	r15 r14 r13 r12 r11 r10 rdx rcx rax
	ret
convert_st0_specials:
; display an FXAM result
	fstp	dword [rdi]			; remove value from FPU stack
	mov	al,ah				; high byte of status
;	and	al,00000111b			; mask C2 to C0 bits
;	and	ah,01000000b			; mask C3 bit
	and	ax,0100000000000111b
	shr	ah,3
	or	al,ah				; combine
	movzx	eax,al
	lea	rcx,[floating_specials]
	mov	rax,qword [rcx + rax*8]
	stosq
	jmp	convert_st0_L9

align 16
convert_float_st0:
; input:	st0 = floating number
;		dl number of mantissa digits to display, must be parity number
;		dh number of exponent digits to display, must be 2,3,4
;		rdi point to output string buffer, must have 4 qwords size

	sub	rsp,8*3

sig_bits	= 57				; number of significant bits in real number display

; convert to decimal
	fstcw	word [rsp + 8*0 + 2]		; save control word
	mov	word [rsp + 8*0 + 4],037Fh	; new control word, round to nearest
	fldcw	word [rsp + 8*0 + 4]		; load control word
	mov	word [rsp + 8*0 + 6],sig_bits

; convert number
	fld	st0
	fxtract					; extract exponent
	fstp	st0				; pop top (significand)
	fisubr	word [rsp + 8*0 + 6]		; subtract bits for significand
	fldl2t					; load log2(10)
	fdivp	st1,st0				; divide, get 10**X
	frndint					; round to nearest
	fist	word [rsp + 8*0 + 6]		; save exponent

; Calculate 10 to the power of ST. Return result in ST. Uses 10**N = 2**(N*Log2(10)).
	fldl2t					; load Log2(10)
	fmulp	st1,st0				; multiply
	call	exp2

	fmulp	st1,st0				; adjust real number for exponent

	neg	word [rsp + 8*0 + 6]		; signed exponent

	fbstp	tbyte [rsp + 8*1]		; save number string

	mov	al,[rsp + 8*1 + 9]		; get the sign byte
	mov	cl,'-'
	bt	eax,7				; check if negative
	mov	al,'+'
	cmovc	eax,ecx
	stosb

	mov	al,[rsp + 8*1 + 8]
	shr	al,4
	add	al,'0'
	stosb
	mov	al,'.'
	stosb
	mov	al,[rsp + 8*1 + 8]
	and	al,00001111b
	add	al,'0'
	stosb

	mov	ecx,8
output_mantissa_L3:
	mov	al,[rsp + 8 + rcx - 1]
	shr	al,4
	add	al,'0'
	stosb
	mov	al,[rsp + 8 + rcx - 1]
	and	al,00001111b
	add	al,'0'
	stosb
	loop	output_mantissa_L3

	movzx	eax,dl
	lea	rdi,[rdi-18 + rax]	; adjust number of digits to display

	mov	eax,'*10^'
	stosd

; exponent sign
	mov	al,'+'
	mov	cx,word [rsp + 8*0 + 6]
	add	cx,17			; adjust for decimal point
	jns	exponent_positive	; jump if not minus
	mov	al,'-'
	neg	cx
exponent_positive:
	stosb
	mov	word [rsp + 8*1],cx
	fild	word [rsp + 8*1]
	fbstp	tbyte [rsp + 8*1]

;	mov	cl,byte [rsp + 8*(3+1+3) + 1]
;	cmp	cl,2
	cmp	dh,2
	jz	exp_display_2
;	cmp	cl,3
	cmp	dh,3
	jz	exp_display_3

exp_display_4:
	mov	al,[rsp + 8*1 + 1]
	shr	al,4
	add	al,'0'
	stosb

exp_display_3:
	mov	al,[rsp + 8*1 + 1]
	and	al,00001111b
	add	al,'0'
	stosb

exp_display_2:
	mov	al,[rsp + 8*1 + 0]
	shr	al,4
	add	al,'0'
	stosb

	mov	al,[rsp + 8*1 + 0]
	and	al,00001111b
	add	al,'0'
	stosb

	fldcw	word [rsp + 8*0 + 2]		; restore control word
	add	rsp,8*3
	ret

align 16
exp2:
; Calculate 2 to the power of ST. Return result in ST(0).
; We needn't to set 03BFh new control word - round to nearest - it has been set already
	fld	st0				; duplicate number
	frndint					; round to integer, ST => W
	fxch					; exchange
	fsub	st0,st1				; get difference, ST => F

; check sign of fractional portion (F)
	ftst					; set flags
	push	rax
	fstsw	word [rsp]			; save flags
	pop	rax
	and	ah,45h				; mask relevant bits
	cmp	ah,1				; check negative bit
	ja	Exp2err				; jump if other bits set, error
	jz	Exp2neg				; jump if negative

; fractional part is positive for
	f2xm1					; ST => (2**F) - 1
	fld1					; load one
	faddp	st1,st0				; ST => 2**F put whole portion (W) on top
	fxch					; put whole portion (W) on top
	fld1					; load one
	fscale					; ST => 2**W
	fmulp	st2,st0				; ST(2) => (2**F) * (2**W)
	fstp	st0				; pop stack
	jmp	Exp2don

; fractional part is negative
Exp2neg:
	fabs					; take absolute value
	f2xm1					; ST => (2**F) - 1
	fld1					; load one
	faddp	st1,st0				; ST => 2**F
	fxch					; put whole portion (W) on top
	fld1					; load one
	fscale					; ST => 2**W
	fdivrp	st2,st0				; ST(2) => (2**W) / (2**F)
	fstp	st0				; pop stack

;finished
Exp2don:
	ret

Exp2err:
	fstp	st0
	fstp	st0				; clear stack
	fld1					; return one
	ret

; -------

align 16
ascii2fpu:
; get mantissa
	sub	rsp,8*3
	mov	word [rsp],00000011b*100h + 7Fh	; 00b Round to nearest (even), 11b 64 bit precision (mantissa)
	fldcw	word [rsp]
	call	get_sign
	mov	qword [rsp+8*2],r11
	call	get_ascii_mantissa
	jc	ascii2fpu_egzyd
	call	format_fp_mantissa
	lea	r15,[rsp+8*0]
	call	pack_number
	jc	ascii2fpu_egzyd
	fbld	tbyte [rsp+8*0]
	cmp	byte [rsp+8*2],0
	jz	skip_change_mantissa_sign
	fchs
skip_change_mantissa_sign:
	lea	rax,[ten_powers]
	lea	rax,[rax + r14*2]
	fld	tbyte [rax + r14*8]		; every uses tbyte = ten bytes
	fmulp	st1,st0

; get exponent
	or	r13,r13				; exponent size
	jz	exponent_done
	lea	rsi,[r12]			; begin of exponent
	mov	r10,r13				; exponent ascii string size
	call	get_sign
	mov	qword [rsp+8*2],r11
	mov	r11,r10				; exponent ascii string size
	lea	r15,[rsp+8*0]
	call	pack_number
	jc	ascii2fpu_egzyd
	fbld	tbyte [rsp+8*0]
	cmp	byte [rsp+8*2],0
	jz	skip_change_exponent_sign
	fchs
skip_change_exponent_sign:
; Calculate 10 to the power of ST. Return result in ST. Uses 10**N = 2**(N*Log2(10)).
	fldl2t					; load Log2(10)
	fmulp	st1,st0				; multiply
	call	exp2

; and terminaly multiply mantissa with 10**exponent
	fmulp	st1,st0
exponent_done:
	add	rsp,8*3
	clc
	ret
ascii2fpu_egzyd:
	add	rsp,8*3
	stc
	ret

align 16
get_sign:
; input:	ascii string at rsi
; return:	process + or - at begin of ascii string, adjust rsi and r10
;		r11=0 if no sign at begin of string or if + at begin of string
;		r11=1 if - at begin of string
;		r12 exponent begin, 0 if no exponent
	push	rax
	mov	al,[rsi]
	cmp	al,'-'
	setz	r11b
	cmp	al,'+'
	setz	al
	or	al,r11b
	movzx	r11d,r11b
	movzx	eax,al
	add	rsi,rax
	sub	r10,rax
	pop	rax
	ret

align 16
get_ascii_mantissa:
; input:	decimal number ascii string at rsi, may contain . as decimal separator, is ended by *10^
;		string size in r10, mantissa must be < 20 (because fbld instruction for load it at the end)
; output:	carry flag if ascii string exceede interval <1,19>
;		r11 mantissa size
;		r12 point to exponent begin
;		r13 number of digits of exponent
	or	r10,r10
	jz	get_ascii_mantissa_bad_size
	mov	r11d,18+1		; 18 digits and 1 decimal pointer .
	cmp	r10,r11
	cmovc	r11,r10
get_ascii_mantissa_L3:
	lea	r13,[rsi+r10]		; point to end of string
	lea	r12,[rsi+r11+4]		; point to exponent begin
	sub	r13,r12			; exponent size
	cmp	dword [rsi+r11],'*10^'
	jz	get_ascii_mantissa_L7
	dec	r11
	jnz	get_ascii_mantissa_L3
	mov	r11d,18+1		; no exponent found, we grab first digits, max 18+1
	cmp	r10,r11
	cmovc	r11,r10
	xor	r13,r13			; zero exponent size
get_ascii_mantissa_L7:
	clc
	ret
get_ascii_mantissa_bad_size:
	stc
	ret

align 16
format_fp_mantissa:
; input:	decimal number ascii string at rsi, may contain . as decimal separator
;		mantissa size in r11
; output:	format string in buffer = remove . in string if it is present
;		r11 string size without .
;		r14 = number_of_digits_after . (correction for exponent)
; destroy:	rsi, destroy string at rsi (move -1 all digits after decimal separator e.g. 123.456 -> 1234566)
	push	rax
	push	rcx
	push	rdi
	mov	al,'.'
	mov	rcx,r11
	lea	rdi,[rsi]
	cld
	repnz scasb
	setnz	al			; al=0 in case of decimal pointer in number, 1 if decimal pointer ins't present
	movzx	eax,al
	mov	r14,rcx
;	jnz	format_fp_mantissa_L5
	push	rsi			; save rsi
	lea	rsi,[rdi]
	lea	rdi,[rdi-1]
	repz movsb
	pop	rsi			; restore rsi
;format_fp_mantissa_L5:
	lea	r11,[r11+rax-1]		; -1 in case of decimal separator in number, don't change r11 in case of absent decimal pointer
	pop	rdi
	pop	rcx
	pop	rax
	ret

align 16
pack_number:
; input:	rsi decimal number ascii string
;		r15 10 byte output buffer
;		r11 string size
; destroy:	r11, r15
	or	r11,r11			; zero string size ?
	jz	pack_number_L9
	cmp	r11,18
	jnbe	pack_number_L9
	push	rax
	xor	eax,eax
	mov	qword [r15],rax		; zeroing output buffer
	mov	dword [r15+6],eax	; zeroing output buffer

align 16
pack_number_L3:
	mov	al,[rsi + r11 - 1]
	sub	al,'0'
	jc	pack_number_L8
	cmp	al,10
	jnc	pack_number_L8
	mov	byte [r15],al
	dec	r11
	jz	pack_number_L7

	mov	al,[rsi + r11 - 1]
	sub	al,'0'
	jc	pack_number_L8
	cmp	al,10
	jnc	pack_number_L8
	shl	al,4
	or	byte [r15],al
	inc	r15
	dec	r11
	jnz	pack_number_L3
pack_number_L7:
	pop	rax
	clc
	ret
pack_number_L8:
	pop	rax
pack_number_L9:
	stc
	ret

align 16
dword_decimal_ascii:
; input:	st0 number
;		rdi buffer for store ascii
; output:	make 0-terminated ascii string at RDI
;		free st0
	push	rax
	push	rcx
	push	rdi
	sub	rsp,8*2
	fbstp	tbyte [rsp]
	movdqa	xmm0,dqword [rsp]
	movdqa	xmm2,dqword [Mask1]
	movdqa	xmm3,dqword [Sum1]
	movdqa	xmm1,xmm0
	psrlq	xmm1,4
	pand	xmm0,xmm2
	pand	xmm1,xmm2
	punpcklbw	xmm0,xmm1
	paddb	xmm0,xmm3
	movdqa	dqword [rsp],xmm0
magz_dygydz_d0_dyzp1ay	=	16
	lea	rdi,[rsp + magz_dygydz_d0_dyzp1ay - 1]
	mov	al,'0'
	mov	ecx,magz_dygydz_d0_dyzp1ay
	std
	repz scasb
	cld
	inc	ecx
; now we have number of nonzero digits
	mov	rdi,qword [rsp + 8*(2+0)]
AMDPad16
dword_decimal_ascii_L0:
	mov	al,byte [rsp + rcx-1]
	stosb
	loop	dword_decimal_ascii_L0
	mov	byte [rdi],cl			; make 0-terminated string
	add	rsp,8*2
	pop	rdi
	pop	rcx
	pop	rax
	ret

align 16
qword_signed_decimal_ascii:
; input:	rax number
;		rdi buffer for store ascii
; output:	make 0-terminated ascii string at RDI
	emms
	push	rax
	push	rcx
	push	rdi
	mov	rcx,rax
	mov	byte [rdi],'+'
	sub	rsp,8*4
	neg	rcx		; rcx = -1 * rax
	or	rax,rax		; is rax negative value ?
	cmovs	rax,rcx
	sets	cl		; set CL=0 if rax positive, CL=1 if rax negative value
	push	rax
	shl	cl,1		; CL*2
	pop	rax
	add	byte [rdi],cl	; change '+' to '-' (2Bh to 2Dh)
	fild	qword [rsp-8]
	fbstp	tbyte [rsp]
	emms
	movdqu	xmm0,dqword [rsp]
	movq	mm0,qword [rsp+8]
	movdqa	xmm2,dqword [Mask1]
	movq	mm2,qword [Mask1]
	movdqa	xmm3,dqword [Sum1]
	movq	mm3,qword [Sum1]
	movdqa	xmm1,xmm0
	movq	mm1,mm0
	psrlq	xmm1,4
	psrld	mm1,4
	pand	xmm0,xmm2
	pand	mm0,mm2
	pand	xmm1,xmm2
	pand	mm1,mm2
	punpcklbw	xmm0,xmm1
	punpcklbw	mm0,mm1
	paddb	xmm0,xmm3
	paddb	mm0,mm3
	movdqu	dqword [rsp],xmm0
	movq	qword [rsp+16],mm0

magz_dygydz_d0_dyzp1ay	=	18
	lea	rdi,[rsp + magz_dygydz_d0_dyzp1ay - 1]
	mov	al,'0'
	mov	ecx,magz_dygydz_d0_dyzp1ay
	std
	repz scasb
	cld
	inc	ecx
; now we have number of nonzero digits
	mov	rdi,qword [rsp + 8*(4+0)]
	inc	rdi			; there is '+' or '-' on string begin
AMDPad16
qword_signed_decimal_ascii_L0:
	mov	al,byte [rsp + rcx-1]
	stosb
	loop	qword_signed_decimal_ascii_L0
	mov	byte [rdi],cl		; make 0-terminated string
	add	rsp,8*4
	pop	rax			; we don't want to pop rdi because rdi points after string
	pop	rcx
	pop	rax
	ret

align 16
bcd2decimalsigned:
; in:	xmm0 holds bcd number
	push	rax rcx rsi
	sub	rsp,8*4
	emms
	movdqu	dqword [rsp],xmm0
	mov	al,'+'
	cld
	stosb
	lea	rsi,[rdi]
	movzx	eax,word [rsp+8]
	test	ah,80h
	setnz	cl
	shl	cl,1
	add	byte [rdi-1],cl	; change '+' to '-' (2Bh to 2Dh)
	and	ah,01111111b
	bsr	ecx,eax
	jz	bcd2decimalsigned_L5
	shr	cl,2
	add	cl,1
; ecx = number of nonzero nybbles
	lea	rdi,[rsp+16]
	call	hexa_64_rax
	lea	rdi,[rsi]
	lea	rsi,[rsp+16+16]
	sub	rsi,rcx
	repz movsb
	lea	rsi,[rdi]
	lea	rdi,[rsp]
	mov	rax,qword [rdi]
	call	hexa_64_rax
	xchg	rsi,rdi
	mov	cl,16
bcd2decimalsigned_L3:
	repz movsb
	mov	ax,'.0'
	stosw
	mov	byte [rdi],cl		; 0-terminator
	add	rsp,8*4
	pop	rsi rcx rax
	ret

bcd2decimalsigned_L5:
	lea	rdi,[rsp]
	mov	rax,qword [rdi]
	call	hexa_64_rax
	xor	edi,edi
	bsr	rcx,rax
	cmovz	rcx,rdi			; at least one digit must be displayed
	shr	cl,2
	add	cl,1
	lea	rdi,[rsi]
	lea	rsi,[rsp+16]
	sub	rsi,rcx
	jmp	bcd2decimalsigned_L3

prepare_clean_floating_st:
	emms
	mov	word [rsp-8],037Fh		; Round to nearest (even),
						; 64-bit precision (mantissa)
	fldcw	word [rsp-8]
	ret
