struc	user
{
	.user_regs	user_regs_struct

	.u_fpvalid	rd	1	; True if math co-processor being used for this mess.
					; Not yet used.
	.pad0		rd	1

	.user_i387	user_i387_struct; Math Co-processor registers.

	.u_tsize	rq	1	; Text segment size (pages).
	.u_dsize	rq	1	; Data segment size (pages).
	.u_ssize	rq	1	; Stack segment size (pages).
	.start_code	rq	1	; Starting virtual address of text.
	.start_stack	rq	1	; Starting virtual address of stack area.
					; This is actually the bottom of the stack,
					; the top of the stack is always found in the
					; rsp register.
	.signal		rq	1	; Signal that caused the core dump.
	.reserved	rd	1	; No longer used
	.pad1		rd	1
	.u_ar0		rq	1	; Used by gdb to help find the values for
					; the registers.
	.u_fpstate	rq	1	; Math Co-processor pointer.
	.magic		rq	1	; To uniquely identify a core file
	.u_comm		rb	32	; User command that was responsible

	.debug_regs	u_debugreg

	.error_code	rq	1	; CPU error code or 0
	.fault_address	rq	1	; CR3 or 0
}


struc	user_regs_struct
{
	.r15		rq	1
	.r14		rq	1
	.r13		rq	1
	.r12		rq	1
	.rbp		rq	1
	.rbx		rq	1
	.r11		rq	1
	.r10		rq	1
	.r9		rq	1
	.r8		rq	1
	.rax		rq	1
	.rcx		rq	1
	.rdx		rq	1
	.rsi		rq	1
	.rdi		rq	1
	.orig_rax	rq	1
	.rip		rq	1
	.cs		rq	1
	.eflags		rq	1
	.rsp		rq	1
	.ss		rq	1
	.fs_base	rq	1
	.gs_base	rq	1
	.ds		rq	1
	.es		rq	1
	.fs		rq	1
	.gs		rq	1
}


struc	user_i387_struct
{
	.cwd		rw	1
	.swd		rw	1
	.twd		rb	1	; Note this is not the same as the 32bit/x87/FSAVE twd
			rb	1	; padding
	.fop		rw	1
	.rip		rq	1
	.rdp		rq	1
	.mxcsr		rd	1
	.mxcsr_mask	rd	1
	.st_space	rb	8*16	; 8*16 bytes for each FP-reg = 128 bytes
	.xmm_space	rb	16*16	; 16*16 bytes for each XMM-reg = 256 bytes
	.padding	rb	8*12
}


struc	u_debugreg
{
	.dr0		rq	1
	.dr1		rq	1
	.dr2		rq	1
	.dr3		rq	1
	.dr4		rq	1	; unused
	.dr5		rq	1	; unused
	.dr6		rq	1
	.dr7		rq	1
; dr8-dr15 unused
}


; Indicate the register numbers for a number of the specific
; debug registers.  Registers 0-3 contain the addresses we wish to trap on.
DR_FIRSTADDR		=	0	; u_debugreg[DR_FIRSTADDR]
DR_LASTADDR		=	3	; u_debugreg[DR_LASTADDR]
DR_STATUS		=	6	; u_debugreg[DR_STATUS]
DR_CONTROL		=	7	; u_debugreg[DR_CONTROL]

; Define a few things for the status register. We can use this to determine
; which debugging register was responsible for the trap. The other bits
; are either reserved or not of interest to us.
DR_TRAP0		=	1	; db0
DR_TRAP1		=	2	; db1
DR_TRAP2		=	4	; db2
DR_TRAP3		=	8	; db3

DR_STEP			=	4000h	; single-step
DR_SWITCH		=	8000h	; task switch

; Now define a bunch of things for manipulating the control register.
; The top two bytes of the control register consist of 4 fields of 4
; bits - each field corresponds to one of the four debug registers,
; and indicates what types of access we trap on, and how large the data
; field is that we are looking at
DR_CONTROL_SHIFT	=	16	; Skip this many bits in ctl register
DR_CONTROL_SIZE		=	4	; 4 control bits per register

DR_RW_EXECUTE		=	0000b	; Settings for the access types to trap on
DR_RW_WRITE		=	0001b
;				0010b	; undefined
DR_RW_READ		=	0011b

DR_LEN_1		=	0000b	; Settings for data length to trap on
DR_LEN_2		=	0100b
DR_LEN_4		=	1100b
DR_LEN_8		=	1000b	; AMD64 only, undefined size in i386

; The low byte to the control register determine which registers are
; enabled.  There are 4 fields of two bits.  One bit is "local", meaning
; that the processor will reset the bit after a task switch and the other
; is global meaning that we have to explicitly reset the bit.  With linux,
; you can use either one, since we explicitly zero the register when we enter
; kernel mode.
DR_LOCAL_ENABLE_SHIFT	=	0	; Extra shift to the local enable bit
DR_GLOBAL_ENABLE_SHIFT	=	1	; Extra shift to the global enable bit
DR_ENABLE_SIZE		=	2	; 2 enable bits per register

DR_LOCAL_ENABLE_MASK	=	01010101b	; Set  local bits for all 4 regs
DR_GLOBAL_ENABLE_MASK	=	10101010b	; Set global bits for all 4 regs

; The second byte to the control register has a few special things.
; We can slow the instruction pipeline for instructions coming via the
; gdt or the ldt if we want to.  I am not sure why this is an advantage
DR_CONTROL_RESERVED	=	0FFFFFFFF0000FC00h	; Reserved
DR_LOCAL_SLOWDOWN	=	0100h	; Local slow the pipeline
DR_GLOBAL_SLOWDOWN	=	0200h	; Global slow the pipeline
